//////////////////////////////////
// Michael Scrivo - 980473300 
// CSC309 A2P2a: OO Concentration
// game written in Javascript
// w/Frames, w/Strategy
//////////////////////////////////

var cardwidth = 80;
var cardheight = 80;

// constants used for the strategy (see playGame() below)
var FOUND = 0;
var NOTFOUND = 1;
var REMOVED = 2;

// autoPlay = 1 tells handleClick that the strategy is playing
// so do not use setTimeout
var autoPlay = 0;	

var CardImages;		// Array containing the images (0 = back)
var WinImages;		// Array containing the winnning images

function rand(num) {
	return Math.floor(Math.random() * num)+1;
}

function loadImages(numFaces) {
	CardImages = new Array(numFaces);
	for (i=0; i<numFaces+1; i++) {
		CardImages[i] = new Image;
		CardImages[i].src = "images/card"+i+".gif";
	}
			
	WinImages = new Array(numFaces);
	for (i=1; i<numFaces+1; i++) {
		WinImages[i] = new Image;
		WinImages[i].src = "images/card"+i+"win.gif";
	}
}

////////////////////////////////
// Board Class & Methods
////////////////////////////////

// class: board
function board(bName, bFrame, numberForMatch, numberOfFaces, numberOfCols) {
	this.name = bName;	// we store the name so it can be called by it
						// in the setTimeout method for example.
	this.frame = bFrame;

	this.numCols = numberOfCols;
	this.numMatch = numberForMatch;
	this.numFaces = numberOfFaces;
	this.numCards = this.numMatch * this.numFaces;
	
	this.cards = new Array(this.numCards);
	
	this.score = 0;		// stores the current # of matches	
	this.state = 0;		// used for determining gameplay options
	this.flipped = 0;	// stores the number of flipped cards
	this.tempFlipped = new Array(this.numMatch);	// stores the indices of flipped cards
	
	// create array of cards and assign their face images
	var count = 0;
	for (i = 0; i < this.numMatch; i++) {
		for (j = 1; j <= this.numFaces; j++) {
			this.cards[count] = new card(this.frame, j);
			count++;
		}
	}
	
	// board methods

	this.drawGrid = drawGridF;
	this.handleClick = handleClickF;
	this.flipBackAll = flipBackAllF;
	this.endGameShow = endGameShowF;
	this.resetGame = resetGameF;
	this.shuffle = shuffleF;
	
	this.playGame = playGameF;
	this.nextUnchecked = nextUncheckedF;
	this.searchSeq = searchSeqF;
}

// board: shuffle()
function shuffleF() {
	// shuffle cards
	for (i=0; i < this.numCards*2; i++) {
		s1 = rand(this.numCards-1);
		s2 = rand(this.numCards-1);
		tempcard = this.cards[s1];
		this.cards[s1] = this.cards[s2];
		this.cards[s2] = tempcard;
	}
	
	// assign the cards a sequence number
	// so we can identify them
	for (i = 0; i < this.numCards; i++)
		this.cards[i].sequenceNum = i;
}

// board: drawGrid()
// Calls the shuffle method and writes the html page containing the game
// cards. Also, draws as many columns as specified by the COLS variable.
function drawGridF() {
    var doc = eval("parent."+this.frame+".document");
	var itemTrack = 0;
	
	this.shuffle();
	
	doc.close();	
	doc.writeln("<html><head></head><body>");
	doc.writeln("	<center><table border=0 cellpadding=0 cellspacing=0>")
	for (i = 0; i < this.numCards; i++) {
		if(itemTrack%this.numCols==0) doc.writeln("		<tr>");
								
		itemTrack++;
		doc.writeln("		<td>")
		doc.writeln("			<a href=\"#\" onClick=\""+"parent.mainFrame."+this.name+".handleClick("+this.cards[i].sequenceNum+"); return false;\">")
		doc.writeln("			<img src=\"images/card0.gif\" width="+cardwidth+" height="+cardheight+" name=\"card"+this.cards[i].sequenceNum+"\" border=0></a>")
		doc.writeln("		</td>")
		
		if (itemTrack%this.numCols==0) {		
			doc.writeln("		<tr>");
			doc.writeln("		</tr>");			
		}
	}
	doc.writeln("		</table></center>")
	doc.writeln("</body></html>");
	doc.close();
}

// board: handleClick(whichCard)
// Handles the onClick event for a the card identified by whichCard.  
// Can handle matching 2 or more cards, depending on the board's numMatch variable.
function handleClickF(whichCard) {
	if (this.state != 0) return;
	
	var cardToFlip = eval("parent."+this.frame+".document.card"+this.cards[whichCard].sequenceNum);
				
	if (cardToFlip.src != CardImages[0].src) return;

	this.cards[whichCard].flip();	// flip card to show it's image
	this.tempFlipped[this.flipped] = whichCard;	// add card to array of flipped cards
	this.flipped++;

	if (this.flipped == this.numMatch) {
		this.state = 1;	// don't allow them to click
		
		// check if they all match, if not, flip them back. We check for matches by comparing 
		// each card in the tempFlipped array to it's neighbour. If any pair does 
		// not match, then it is not a match, and we flip back.
		for (i = 0; i < this.numMatch-1; i++) {
			if (!this.cards[this.tempFlipped[i]].equals(this.cards[this.tempFlipped[i+1]])) {
				if (autoPlay == 0) 
					setTimeout(this.name+".flipBackAll()",500);
				else
					this.flipBackAll();
					
				return;
			}
		}
		
		// if we get here, everything matches.
		// Reset state, and clear flipped cards.
		this.state = 0;
		this.flipped = 0;		
		this.tempFlipped = null;
		this.tempFlipped = new Array(this.numMatch);

		this.score++;  // increment number of matches found
		
		// when the score equals the number of faces. the player
		// has won, so initiate the end game sequence.
		if (this.score == this.numFaces) {
			this.state = 1;		// don't allow them to click
			
			this.endGameShow();
			gameOver = 1;
		}
	}
}

// board: flipBackAll()
// flips all cards back to their covered state.
// resets board variables.
function flipBackAllF() {
	for (i=0; i < this.tempFlipped.length; i++) 
	{
		this.cards[this.tempFlipped[i]].flipBack();
	}
	this.state = 0;
	this.flipped = 0;		
	this.tempFlipped = null;
	this.tempFlipped = new Array(this.numMatch);
}

// board: endGameShow()
// Displays animation, by alternating the card image with it's 
// winning image for a fixed amount of time. Then resets the board.
function endGameShowF() {
	this.state++;
	
	if (this.state > 20) {
		return this.resetGame();
	}
	
	for (i = 0; i < this.numCards; i++) {
		this.cards[i].flipWin(this.state%2);
	}
	
	if(autoPlay == 0)
		setTimeout(this.name+".endGameShow()",50)
	else
		this.endGameShow();	
			
	return true;
}

// board: resetGame() 
// resets board variables, flips cards to covered state and shuffles
function resetGameF() {
	this.state = 0;
	this.score = 0;
	this.flipped = 0;
	this.tempFlipped = null;
	this.tempFlipped = new Array(this.numMatch);
	
	for (i = 0; i < this.numCards; i++)
		this.cards[i].flipBack();
			
	this.shuffle();
}

// board: playGameF(numGames)
// the program proceeds to play 'numGames' games using the coded strategy
// and collect the results
function playGameF(numGames) {

	if (this.numMatch != 2) {
		alert("Sorry, you can only play the strategy with number for match = 2");
		return;
	}

	// stores the status of each card on the board, either
	// FOUND - card has been seen, but not matched
	// NOTFOUND - card has not been seen
	// REMOVED - card has been matched
	var cardStatus = new Array(this.numCards);
	
	// newmatch represents a newly discovered match on the prevous turn.
	// It is the second card picked on the previous turn, who's matching card 
	// we've seen before, but because it was the second card picked, 
	// we couldnt choose it during that turn.
	var newMatch = new Array(2);
	
	// used for satistical analysis
	var games = 0; count=0, worstCase=0, total = 0, average=0;
	
	gameOver=0
	
	// notifies the handleClick function that the strategy is playing
	// and that it should not use the setTimeout function.
	autoPlay = 1;	
			
	while (games != numGames) {
		
		newMatch[0] = null;
		
		// new game:  no cards have been seen
		for (j=0; j<this.numCards; j++)
			cardStatus[j] = NOTFOUND; 

		// keep playing until all matches have been found
		while (gameOver!= 1) {
			
			// newmatch holds a card, so we know of a match.
			// so we select the two cards
			if (newMatch[0] != null) {
				this.handleClick(newMatch[0]);
				this.handleClick(this.searchSeq(cardStatus, newMatch[0], newMatch[1]));
				count+=2;
				newMatch[0] = null;
			}
			// newmatch is empty so we don't know of any matching pairs
			else {
				// flips next unchecked card, to gain new information
				var a = this.nextUnchecked (cardStatus);
				cardStatus[a] = FOUND;
				this.handleClick(a);
				count++;
				
				var y = this.searchSeq(cardStatus,a,this.cards[a].image.src);
				
				// if a match for 'a' is found, we select the match
				if (y != -1) {
					this.handleClick(y);
					count++;
					cardStatus[a] = REMOVED;
					cardStatus[y] = REMOVED;
				}
				// no match for 'a' has been previously seen
				else
				{
					// since we dont know where 'a' match is, we know
					// it's in the unchecked cards, so we pick from them
					// to gain new information
					var b = this.nextUnchecked(cardStatus);
					this.handleClick(b);
					count++;
					cardStatus[b] = FOUND;
					
					// since b was unchecked it may now match a.
					if(this.cards[a].image.src  == this.cards[b].image.src) {
						cardStatus[a] = REMOVED;
						cardStatus[b] = REMOVED;
					}
					else {
						// we may have previously seen a match for b, if so, well pick b and
						// its match on the next turn
						if (this.searchSeq(cardStatus,b,this.cards[b].image.src) != -1) {
							newMatch[0] = b;
							newMatch[1] = this.cards[b].image.src;
						}
						// else no matches found or known about
					}
				}
			}	
		}
		
		if (count > worstCase)
			worstCase = count;

		total+=count;
		count = 0;
		
		gameOver = 0;	// reset game
		games++;
	}
	
	average = total/numGames;
	autoPlay = 0;
	alert("Worst Case: " + worstCase + " flips\nAverage: " + average + " flips");
}

// board: nextUncheckedF(array)
// returns the number in the sequence of the next card which has not been flipped
function nextUncheckedF(array) {
	for (i=0; i< this.numCards; i++) {
		if (array[i] == NOTFOUND)
			return i;
	}
	return -1;
}

// board: searchSeqF(array, seq, src)
// returns the number in the sequence of a card that match the card at 'seq'
// return -1 is no match found
function searchSeqF(array, seq, src) {
	for (i=0; i<this.numCards; i++) {
		if (src == this.cards[i].image.src && i != seq && array[i] == FOUND)
			return i;
	}
	
	return -1;
}

////////////////////////////////
// Card Class & Methods
////////////////////////////////

// class: card
function card(bFrame, faceNum) {
	this.frame = bFrame;
	this.face = faceNum;	// used to identify correpsonding winning image
	this.image = new Image;
	this.image.src = CardImages[this.face].src;		// card's face

	// Methods
	this.equals = equalsF;
	this.flip = flipF;
	this.flipBack = flipBackF;
	this.flipWin = flipWinF;
}

// card: equals(card)
// returns true if the card passed in has the same
// image.src as the current card.
function equalsF(card) {
	if (card.image.src == this.image.src) 
		return true;
	else
		return false;
}

// card: flipBack()
// Flips the card back to it's covered state
function flipBackF() {
	var card = eval("parent."+this.frame+".document.card"+this.sequenceNum);
	card.src = CardImages[0].src;
}

// card: flip()
// Flips the card to it's face up state.
function flipF() {
	var card = eval("parent."+this.frame+".document.card"+this.sequenceNum); 
	card.src = this.image.src;
}

// card: flipWin(state)
// Flips the card between it's winning image 
// and it's normal image based on the state value.
function flipWinF(state) {
	var card = eval("parent."+this.frame+".document.card"+this.sequenceNum); 
	if (state == 1)
		card.src = this.image.src;
	else
		card.src = WinImages[this.face].src;
}